@@ -1,5 +1,5 @@
 /*
-	SPDX-FileCopyrightText: 2008-2021 Graeme Gott <graeme@gottcode.org>
+	SPDX-FileCopyrightText: 2008-2024 Graeme Gott <graeme@gottcode.org>
 
 	SPDX-License-Identifier: GPL-3.0-or-later
 */
@@ -7,6 +7,7 @@
 #include "board.h"
 
 #include "appearance_dialog.h"
+#include "edge_scroller.h"
 #include "generator.h"
 #include "message.h"
 #include "overview.h"
@@ -19,70 +20,36 @@
 #include <QDir>
 #include <QFile>
 #include <QFileInfo>
+#include <QGridLayout>
 #include <QImageReader>
-#include <QMatrix4x4>
 #include <QMessageBox>
 #include <QMouseEvent>
-#include <QOpenGLTexture>
 #include <QPainter>
 #include <QSettings>
-#include <QWheelEvent>
+#include <QTimeLine>
 #include <QVector2D>
+#include <QWheelEvent>
 #include <QXmlStreamReader>
 #include <QXmlStreamWriter>
 
 #include <algorithm>
-
 #include <cmath>
-#include <cstdlib>
-
-//-----------------------------------------------------------------------------
-
-int powerOfTwo(int value)
-{
-	value--;
-	value |= value >> 1;
-	value |= value >> 2;
-	value |= value >> 4;
-	value |= value >> 8;
-	value |= value >> 16;
-	value++;
-	return value;
-}
-
-namespace
-{
-
-struct PieceDetails
-{
-	QPoint pos;
-	int rotation;
-	QList<Tile*> tiles;
-
-	PieceDetails(const QPoint& pos_, int rotation_, const QList<Tile*>& tiles_)
-		: pos(pos_), rotation(rotation_), tiles(tiles_)
-	{
-	}
-};
-
-}
 
 //-----------------------------------------------------------------------------
 
 Board::Board(QWidget* parent)
-	: QOpenGLWidget(parent)
+	: QWidget(parent)
 	, m_id(0)
 	, m_load_bevels(true)
 	, m_has_bevels(true)
 	, m_has_shadows(true)
-	, m_image(nullptr)
-	, m_image_ts(0)
+	, m_bevel_pixmap(":/bumpmap.png")
 	, m_columns(0)
 	, m_rows(0)
 	, m_total_pieces(0)
 	, m_completed(0)
 	, m_pos(0, 0)
-	, m_scale_level(9)
+	, m_scale_level(-1)
 	, m_scale(0)
 	, m_scrolling(false)
 	, m_selecting(false)
@@ -94,12 +61,62 @@ Board::Board(QWidget* parent)
 	setFocusPolicy(Qt::StrongFocus);
 	setFocus();
 	setMouseTracking(true);
+	setAutoFillBackground(true);
 
 	m_message = new Message(this);
 
+	// Animate zooming to best fit
+	m_zoom_timer = new QTimeLine(250, this);
+	m_zoom_timer->setEasingCurve(QEasingCurve::Linear);
+	m_zoom_timer->setFrameRange(0, 10);
+	m_zoom_timer->setUpdateInterval(25);
+	connect(m_zoom_timer, &QTimeLine::frameChanged, this, [this](int value) {
+		zoom(value, false);
+	});
+
+	// Create edge scrollers
+	m_scroll_left = new EdgeScroller(1, 0, this);
+	m_scroll_left->hide();
+	connect(m_scroll_left, &EdgeScroller::edgeScroll, this, &Board::edgeScroll);
+
+	m_scroll_right = new EdgeScroller(-1, 0, this);
+	m_scroll_right->hide();
+	connect(m_scroll_right, &EdgeScroller::edgeScroll, this, &Board::edgeScroll);
+
+	m_scroll_up = new EdgeScroller(0, 1, this);
+	m_scroll_up->hide();
+	connect(m_scroll_up, &EdgeScroller::edgeScroll, this, &Board::edgeScroll);
+
+	m_scroll_down = new EdgeScroller(0, -1, this);
+	m_scroll_down->hide();
+	connect(m_scroll_down, &EdgeScroller::edgeScroll, this, &Board::edgeScroll);
+
+	QGridLayout* layout = new QGridLayout(this);
+
+	layout->setColumnMinimumWidth(0, 16);
+	layout->setColumnStretch(1, 1);
+	layout->setColumnMinimumWidth(2, 16);
+
+	layout->setRowMinimumHeight(0, 16);
+	layout->setRowStretch(1, 1);
+	layout->setRowMinimumHeight(2, 16);
+
+	layout->setSpacing(0);
+	layout->setContentsMargins(0, 0, 0, 0);
+
+	layout->addWidget(m_scroll_left, 1, 0);
+	layout->addWidget(m_scroll_right, 1, 2);
+	layout->addWidget(m_scroll_up, 0, 1);
+	layout->addWidget(m_scroll_down, 2, 1);
+
 	// Create overview dialog
 	m_overview = new Overview(parent);
 	connect(m_overview, &Overview::toggled, this, &Board::overviewToggled);
+
+	// Load colors
+	AppearanceDialog dialog;
+	dialog.accept();
+	setAppearance(dialog);
 }
 
 //-----------------------------------------------------------------------------
@@ -107,16 +124,12 @@ Board::Board(QWidget* parent)
 Board::~Board()
 {
 	cleanup();
-	delete m_bumpmap_image;
-	delete m_shadow_image;
 	delete m_message;
-	delete graphics_layer;
-	graphics_layer = nullptr;
 }
 
 //-----------------------------------------------------------------------------
 
-Piece* Board::findCollidingPiece(Piece* piece) const
+Piece* Board::findCollidingPiece(const Piece* piece) const
 {
 	for (int i = m_pieces.count() - 1; i >= 0; --i) {
 		Piece* other = m_pieces.at(i);
@@ -124,7 +137,7 @@ Piece* Board::findCollidingPiece(Piece* piece) const
 			return other;
 		}
 	}
-	return 0;
+	return nullptr;
 }
 
 //-----------------------------------------------------------------------------
@@ -140,26 +153,28 @@ void Board::removePiece(Piece* piece)
 
 void Board::setAppearance(const AppearanceDialog& dialog)
 {
-	makeCurrent();
-
 	m_has_bevels = dialog.hasBevels();
 	m_has_shadows = dialog.hasShadows();
 
 	QPalette palette = dialog.colors();
-	graphics_layer->setClearColor(palette.color(QPalette::Base).darker(150));
+	palette.setColor(backgroundRole(), palette.color(QPalette::Base).darker(150));
 	setPalette(palette);
 	for (Piece* piece : std::as_const(m_pieces)) {
 		piece->setSelected(piece->isSelected());
 	}
+
+	const qreal pixelratio = devicePixelRatioF();
+	m_shadow_pixmap = dialog.shadow(pixelratio);
+	m_selected_shadow_pixmap = dialog.shadowSelected(pixelratio);
 }
 
 //-----------------------------------------------------------------------------
 
-void Board::updateSceneRectangle(Piece* piece)
+void Board::updateSceneRectangle(const Piece* piece)
 {
-	int size = Tile::size / 2;
+	const int size = Tile::size / 2;
 	m_scene = m_scene.united(piece->boundingRect().adjusted(-size, -size, size, size));
-	updateArray(m_scene_array, m_scene, 0);
+	updateViewport();
 }
 
 //-----------------------------------------------------------------------------
@@ -168,7 +183,7 @@ void Board::newGame(const QString& image, int difficulty)
 {
 	// Remove any previous textures and tiles
 	cleanup();
-	zoom(0);
+	zoom(0, false);
 
 	// Prevent starting a game with a missing image
 	if (!QFileInfo::exists(Path::image(image))) {
@@ -181,6 +196,7 @@ void Board::newGame(const QString& image, int difficulty)
 	QApplication::setOverrideCursor(Qt::WaitCursor);
 	m_message->setText(tr("Please Wait"));
 	m_message->setVisible(true);
+	QCoreApplication::processEvents();
 
 	// Generate ID
 	m_id = 0;
@@ -191,7 +207,7 @@ void Board::newGame(const QString& image, int difficulty)
 	m_id++;
 
 	// Find puzzle dimensions
-	QSizeF size = QImageReader(Path::image(image)).size();
+	const QSizeF size = QImageReader(Path::image(image)).size();
 	if (size.width() > size.height()) {
 		m_columns = 4 * difficulty;
 		m_rows = std::max(std::lround(m_columns * size.height() / size.width()), 1L);
@@ -209,19 +225,19 @@ void Board::newGame(const QString& image, int difficulty)
 
 	// Generate puzzle
 	updateStatusMessage(tr("Generating puzzle..."));
-	Generator generator(m_columns, m_rows, m_random);
+	const Generator generator(m_columns, m_rows, m_random);
 	QList<QList<Tile*>> pieces = generator.pieces();
 	std::shuffle(pieces.begin(), pieces.end(), m_random);
 
 	updateStatusMessage(tr("Creating pieces..."));
-	int count = pieces.count();
-	int step = (count > 25) ? (count / 25) : 1;
+	const int count = pieces.count();
+	const int step = (count > 25) ? (count / 25) : 1;
 	for (int i = 0; i < count; ++i) {
 		// Create piece
 		Piece* piece = new Piece(QPoint(0, 0), randomInt(4), pieces.at(i), this);
 		m_pieces.append(piece);
 		piece->setPosition(m_pos - QRect(QPoint(0,0), piece->boundingRect().size()).center());
-		piece->pushNeighbors();
+		piece->pushCollidingPieces();
 
 		// Show pieces
 		if ((i % step) == 0) {
@@ -231,8 +247,8 @@ void Board::newGame(const QString& image, int difficulty)
 		}
 	}
 
-	for (int i = 0; i < count; ++i) {
-		m_pieces.at(i)->findNeighbors(m_pieces);
+	for (Piece* piece : std::as_const(m_pieces)) {
+		piece->findSolutionNeighbors(m_pieces);
 	}
 	Q_EMIT clearMessage();
 
@@ -256,6 +272,7 @@ void Board::openGame(int id)
 	QApplication::setOverrideCursor(Qt::WaitCursor);
 	m_message->setText(tr("Please Wait"));
 	m_message->setVisible(true);
+	QCoreApplication::processEvents();
 
 	// Open saved game file
 	QFile file(Path::save(id));
@@ -273,7 +290,7 @@ void Board::openGame(int id)
 	QXmlStreamAttributes attributes = xml.attributes();
 	int board_zoom = 0;
 	QRect rect;
-	unsigned int version = attributes.value("version").toUInt();
+	const unsigned int version = attributes.value("version").toUInt();
 	if (xml.name() == QLatin1String("tetzle") && version <= 5) {
 		m_image_path = attributes.value("image").toString();
 		if (!QFileInfo::exists(Path::image(m_image_path))) {
@@ -282,20 +299,23 @@ void Board::openGame(int id)
 			cleanup();
 			return;
 		}
+
 		board_zoom = attributes.value("zoom").toInt();
+
 		m_pos.setX(attributes.value("x").toInt());
 		m_pos.setY(attributes.value("y").toInt());
-		QStringList values = attributes.value("rect").toString().split(",");
+
+		const QStringList values = attributes.value("rect").toString().split(",");
 		rect.setRect(values.value(0).toInt(),
-			values.value(1). toInt(),
-			values.value(2).toInt(),
-			values.value(3).toInt());
+				values.value(1).toInt(),
+				values.value(2).toInt(),
+				values.value(3).toInt());
 	} else {
 		xml.raiseError(tr("Unknown data format"));
 	}
 
-	// Load piece details
-	QList<PieceDetails> pieces;
+	// Load pieces
+	updateStatusMessage(tr("Loading pieces..."));
 	QPoint pos;
 	int rotation = -1;
 	QList<Tile*> tiles;
@@ -306,7 +326,8 @@ void Board::openGame(int id)
 		xml.readNext();
 
 		if (xml.isEndElement() && (xml.name() == QLatin1String("piece") || xml.name() == QLatin1String("group"))) {
-			pieces.append(PieceDetails(pos, rotation, tiles));
+			m_pieces.append(new Piece(pos, rotation, tiles, this));
+			tiles.clear();
 		}
 		if (!xml.isStartElement()) {
 			continue;
@@ -319,14 +340,18 @@ void Board::openGame(int id)
 				pos = QPoint(attributes.value("x").toInt(), attributes.value("y").toInt());
 				rotation = (rotation != -1) ? rotation : attributes.value("rotation").toInt();
 			}
-			int column = attributes.value("column").toInt();
+
+			const int column = attributes.value("column").toInt();
 			m_columns = std::max(m_columns, column);
-			int row = attributes.value("row").toInt();
+			const int row = attributes.value("row").toInt();
 			m_rows = std::max(m_rows, row);
-			int bevel = attributes.value("bevel").toInt();
+
+			// Old games converted to new games won't have bevels
+			const int bevel = attributes.value("bevel").toInt();
 			if (bevel) {
 				m_load_bevels = true;
 			}
+
 			Tile* tile = new Tile(column, row);
 			tile->setBevel(bevel);
 			tiles.append(tile);
@@ -334,12 +359,10 @@ void Board::openGame(int id)
 			attributes = xml.attributes();
 			pos = QPoint(attributes.value("x").toInt(), attributes.value("y").toInt());
 			rotation = attributes.value("rotation").toInt();
-			tiles.clear();
 		} else if (xml.name() == QLatin1String("group")) {
 			piece = false;
-			auto r = xml.attributes().value("rotation");
+			const QStringView r = xml.attributes().value("rotation");
 			rotation = !r.isEmpty() ? r.toInt() : -1;
-			tiles.clear();
 		} else if (xml.name() != QLatin1String("overview")) {
 			xml.raiseError(tr("Unknown element '%1'").arg(xml.name().toString()));
 		}
@@ -352,35 +375,28 @@ void Board::openGame(int id)
 	}
 	m_total_pieces = (++m_columns * ++m_rows) / 4;
 
+	for (Piece* piece : std::as_const(m_pieces)) {
+		piece->findSolutionNeighbors(m_pieces);
+	}
+
 	// Load image
 	updateStatusMessage(tr("Loading image..."));
 	loadImage();
-
-	// Load pieces
-	updateStatusMessage(tr("Loading pieces..."));
-	int count = pieces.count();
-	for (int i = 0; i < count; ++i) {
-		const PieceDetails& details = pieces.at(i);
-		m_pieces.append( new Piece(details.pos, details.rotation, details.tiles, this) );
-	}
-	for (int i = 0; i < count; ++i) {
-		m_pieces.at(i)->findNeighbors(m_pieces);
-	}
 	Q_EMIT clearMessage();
 
 	// Load scene rectangle
 	updateSceneRectangle();
 	if (rect.contains(m_scene)) {
 		m_scene = rect;
-		updateArray(m_scene_array, m_scene, 0);
 	}
+	updateViewport();
 
 	// Draw tiles
 	m_message->setVisible(false);
 	if (version > 3) {
-		zoom(board_zoom);
+		zoom(board_zoom, false);
 	} else {
-		zoom(0);
+		zoom(0, false);
 		retrievePieces();
 	}
 	QApplication::restoreOverrideCursor();
@@ -390,7 +406,7 @@ void Board::openGame(int id)
 
 //-----------------------------------------------------------------------------
 
-void Board::saveGame()
+void Board::saveGame() const
 {
 	if (pieceCount() <= 1) {
 		return;
@@ -413,25 +429,27 @@ void Board::saveGame()
 	xml.writeAttribute("zoom", QString::number(m_scale_level));
 	xml.writeAttribute("x", QString::number(m_pos.x()));
 	xml.writeAttribute("y", QString::number(m_pos.y()));
-	xml.writeAttribute("rect", QString("%1,%2,%3,%4").
-		arg(m_scene.x())
-		.arg(m_scene.y())
-		.arg(m_scene.width())
-		.arg(m_scene.height()));
+	xml.writeAttribute("rect", QString("%1,%2,%3,%4")
+			.arg(m_scene.x())
+			.arg(m_scene.y())
+			.arg(m_scene.width())
+			.arg(m_scene.height()));
 
-	for (Piece* piece : std::as_const(m_pieces)) {
+	for (const Piece* piece : std::as_const(m_pieces)) {
 		piece->save(xml);
 	}
-	for (Piece* piece : std::as_const(m_selected_pieces)) {
+	for (const Piece* piece : std::as_const(m_selected_pieces)) {
 		piece->save(xml);
 	}
-	for (Piece* piece : std::as_const(m_active_pieces)) {
+	for (const Piece* piece : std::as_const(m_active_pieces)) {
 		piece->save(xml);
 	}
 
 	xml.writeEndElement();
 
 	xml.writeEndDocument();
+
+	QSettings().setValue("Overview/Visible", m_overview->isVisible());
 }
 
 //-----------------------------------------------------------------------------
@@ -441,6 +459,7 @@ void Board::retrievePieces()
 	// Inform user this will take awhile
 	updateStatusMessage(tr("Retrieving pieces..."));
 	QApplication::setOverrideCursor(Qt::WaitCursor);
+	QCoreApplication::processEvents();
 
 	// Make sure all pieces are free
 	QList<Piece*> pieces = m_pieces + m_active_pieces + m_selected_pieces;
@@ -458,7 +477,7 @@ void Board::retrievePieces()
 		m_pieces.append(piece);
 		piece->setPosition(m_pos - QRect(QPoint(0,0), piece->boundingRect().size()).center());
 		piece->setSelected(false);
-		piece->pushNeighbors();
+		piece->pushCollidingPieces();
 	}
 
 	// Update view
@@ -473,14 +492,14 @@ void Board::retrievePieces()
 
 void Board::zoomIn()
 {
-	zoom(m_scale_level + 1);
+	zoom(m_scale_level + 1, true);
 }
 
 //-----------------------------------------------------------------------------
 
 void Board::zoomOut()
 {
-	zoom(m_scale_level - 1);
+	zoom(m_scale_level - 1, true);
 }
 
 //-----------------------------------------------------------------------------
@@ -488,11 +507,11 @@ void Board::zoomOut()
 void Board::zoomFit()
 {
 	// Find new scale level
-	float sx = static_cast<float>(width()) / static_cast<float>(m_scene.width());
-	float sy = static_cast<float>(height()) / static_cast<float>(m_scene.height());
-	float factor = qBound(0.0f, std::min(sx, sy), 1.0f);
+	const float sx = static_cast<float>(width()) / static_cast<float>(m_scene.width());
+	const float sy = static_cast<float>(height()) / static_cast<float>(m_scene.height());
+	const float factor = std::clamp(std::min(sx, sy), 0.0f, 1.0f);
 	int level = 0;
-	for (int i = 9; i >= 0; --i) {
+	for (int i = ZoomSlider::maxScaleLevel(); i >= 0; --i) {
 		if (ZoomSlider::scaleFactor(i) <= factor) {
 			level = i;
 			break;
@@ -501,48 +520,58 @@ void Board::zoomFit()
 
 	m_pos = m_scene.center();
 	if (m_scale_level == level) {
+		updateViewport();
 		update();
 		return;
 	}
 
 	// Animate zoom
-	int delta = (level > m_scale_level) ? 1 : -1;
-	int count = abs(level - m_scale_level);
-	for (int i = 0; i < count; ++i) {
-		zoom(m_scale_level + delta);
+	const int count = abs(level - m_scale_level);
+	if (level > m_scale_level) {
+		m_zoom_timer->setDirection(QTimeLine::Forward);
+		m_zoom_timer->setFrameRange(m_scale_level, level);
+	} else {
+		m_zoom_timer->setDirection(QTimeLine::Backward);
+		m_zoom_timer->setFrameRange(level, m_scale_level);
 	}
+	m_zoom_timer->setDuration(count * m_zoom_timer->updateInterval());
+	m_zoom_timer->start();
 }
 
 //-----------------------------------------------------------------------------
 
-void Board::zoom(int level)
+void Board::zoom(int level, bool on_cursor)
 {
-	QPoint old_pos = mapCursorPosition();
+	level = std::clamp(level, 0, ZoomSlider::maxScaleLevel());
+	if (level == m_scale_level) {
+		return;
+	}
+
+	// Find cursor position
+	const QPoint cursor_pos = mapFromGlobal(QCursor::pos());
+	const QPoint old_pos = mapPosition(cursor_pos);
 
 	// Calculate new scale value
-	m_scale_level = qBound(0, level, 9);
+	m_scale_level = level;
 	m_scale = ZoomSlider::scaleFactor(m_scale_level);
 
-	// Update mouse cursor position
-	QPoint new_pos = mapCursorPosition();
-	int count = m_active_pieces.count();
-	for (int i = 0; i < count; ++i) {
-		m_active_pieces.at(i)->moveBy(new_pos - old_pos);
+	// Center zoom on cursor
+	if (on_cursor && rect().contains(cursor_pos)) {
+		m_pos += old_pos - mapPosition(cursor_pos);
 	}
-	updateCursor();
 
 	// Update scene
+	updateViewport();
 	update();
-	Q_EMIT zoomChanged(m_scale_level, m_scale);
-	Q_EMIT zoomOutAvailable(m_scale_level > 0);
-	Q_EMIT zoomInAvailable(m_scale_level < 9);
+	updateCursor();
+	Q_EMIT zoomChanged(m_scale_level);
 }
 
 //-----------------------------------------------------------------------------
 
 void Board::toggleOverview()
 {
-	bool visible = !m_overview->isVisible();
+	const bool visible = !m_overview->isVisible();
 	m_overview->setVisible(visible);
 	if (visible) {
 		activateWindow();
@@ -551,151 +580,101 @@ void Board::toggleOverview()
 }
 
 //-----------------------------------------------------------------------------
-
-void Board::initializeGL()
+#if QT_VERSION >= QT_VERSION_CHECK(6, 6, 0)
+bool Board::event(QEvent* event)
 {
-	// Configure OpenGL
-	GraphicsLayer::init();
-
-	// Load static images
-	m_bumpmap_image = new QOpenGLTexture(QImage(":/bumpmap.png"));
-	m_bumpmap_image->setMinificationFilter(QOpenGLTexture::LinearMipMapLinear);
-	m_bumpmap_image->setMagnificationFilter(QOpenGLTexture::Linear);
-
-	m_shadow_image = new QOpenGLTexture(QImage(":/shadow.png"));
-	m_shadow_image->setMinificationFilter(QOpenGLTexture::LinearMipMapLinear);
-	m_shadow_image->setMagnificationFilter(QOpenGLTexture::Linear);
-
-	// Load colors
-	AppearanceDialog dialog;
-	dialog.accept();
-	setAppearance(dialog);
+	if (event->type() == QEvent::DevicePixelRatioChange) {
+		loadImage();
+		for (Piece* piece : std::as_const(m_pieces)) {
+			piece->setPosition(piece->scenePos());
+		}
+	}
+	return QWidget::event(event);
 }
-
+#endif
 //-----------------------------------------------------------------------------
 
-void Board::resizeGL(int w, int h)
+void Board::paintEvent(QPaintEvent*)
 {
-	const qreal pixelratio = devicePixelRatioF();
-
-	graphics_layer->setViewport(0, 0, w * pixelratio, h * pixelratio);
-
-	QMatrix4x4 matrix;
-	matrix.ortho(0, w * pixelratio, h * pixelratio, 0, -4000, 3);
-	graphics_layer->setProjection(matrix);
-
-	m_message->setViewport(size());
-}
-
-//-----------------------------------------------------------------------------
-
-void Board::paintGL()
-{
-	graphics_layer->clear();
-
-	graphics_layer->uploadData();
+	QPainter painter(this);
+	painter.setRenderHint(QPainter::SmoothPixmapTransform);
 
 	// Transform viewport
-	const qreal pixelratio = devicePixelRatioF();
-	QRect viewport = rect();
-	viewport.setSize(viewport.size() * pixelratio);
-	QMatrix4x4 matrix;
-	matrix.scale(m_scale * pixelratio, m_scale * pixelratio);
-	matrix.translate((width() / (2 * m_scale)) - m_pos.x(), (height() / (2 * m_scale)) - m_pos.y());
-	graphics_layer->setModelview(matrix);
+	painter.save();
+	painter.setTransform(m_viewport_transform);
 
 	// Draw scene rectangle
 	QColor fill = palette().color(QPalette::Base);
 	QColor border = fill.lighter(125);
 	if (m_scene.isValid()) {
-		drawArray(m_scene_array, fill, border);
+		drawRect(painter, m_scene, fill, border);
 	}
 
 	// Draw pieces
-	if (m_image) {
-		graphics_layer->bindTexture(0, m_image->textureId());
+	if (!m_pixmap.isNull()) {
+		FragmentList bevel;
+		FragmentList shadow;
+		FragmentList tiles;
+		const int total_tiles = m_columns * m_rows;
+		bevel.reserve(total_tiles);
+		shadow.reserve(total_tiles);
+		tiles.reserve(total_tiles);
+		for (const Piece* piece : std::as_const(m_pieces)) {
+			const QRect r = m_viewport_transform.mapRect(piece->boundingRect());
+			if (m_viewport.intersects(r)) {
+				shadow.append(piece->shadow());
+				tiles.append(piece->tiles());
+				bevel.append(piece->bevel());
+			}
+		}
+		if (m_has_shadows) {
+			shadow.draw(painter, m_shadow_pixmap);
+		}
+		tiles.draw(painter, m_pixmap, QPainter::OpaqueHint);
 		if (m_has_bevels && m_load_bevels) {
-			graphics_layer->setTextureUnits(2);
-			graphics_layer->bindTexture(1, m_bumpmap_image->textureId());
+			bevel.draw(painter, m_bevel_pixmap);
 		}
 
-		int count = m_pieces.count();
-		for (int i = 0; i < count; ++i) {
-			QRect r = matrix.mapRect(m_pieces.at(i)->boundingRect());
-			if (viewport.intersects(r)) {
-				m_pieces.at(i)->drawTiles();
+		for (const Piece* piece : std::as_const(m_selected_pieces)) {
+			if (m_has_shadows) {
+				piece->shadow().draw(painter, m_selected_shadow_pixmap);
+			}
+			piece->tiles().draw(painter, m_pixmap, QPainter::OpaqueHint);
+			if (m_has_bevels && m_load_bevels) {
+				piece->bevel().draw(painter, m_bevel_pixmap);
 			}
 		}
 
-		count = m_selected_pieces.count();
-		for (int i = 0; i < count; ++i) {
-			m_selected_pieces.at(i)->drawTiles();
-		}
-
-		count = m_active_pieces.count();
-		for (int i = 0; i < count; ++i) {
-			m_active_pieces.at(i)->drawTiles();
-		}
-
-		if (m_has_bevels) {
-			graphics_layer->setTextureUnits(1);
-		}
-	}
-
-	// Draw shadows
-	graphics_layer->setBlended(true);
-	if (m_image && m_has_shadows) {
-		graphics_layer->bindTexture(0, m_shadow_image->textureId());
-
-		graphics_layer->setColor(palette().color(QPalette::Text));
-		int count = m_pieces.count();
-		for (int i = 0; i < count; ++i) {
-			QRect r = matrix.mapRect(m_pieces.at(i)->boundingRect());
-			if (viewport.intersects(r)) {
-				m_pieces.at(i)->drawShadow();
+		for (const Piece* piece : std::as_const(m_active_pieces)) {
+			if (m_has_shadows) {
+				piece->shadow().draw(painter, m_selected_shadow_pixmap);
+			}
+			piece->tiles().draw(painter, m_pixmap, QPainter::OpaqueHint);
+			if (m_has_bevels && m_load_bevels) {
+				piece->bevel().draw(painter, m_bevel_pixmap);
 			}
 		}
-
-		graphics_layer->setColor(palette().color(QPalette::Highlight));
-		count = m_selected_pieces.count();
-		for (int i = 0; i < count; ++i) {
-			m_selected_pieces.at(i)->drawShadow();
-		}
-
-		count = m_active_pieces.count();
-		for (int i = 0; i < count; ++i) {
-			m_active_pieces.at(i)->drawShadow();
-		}
-
-		graphics_layer->setColor(Qt::white);
 	}
 
 	// Untransform viewport
-	if (qFuzzyCompare(pixelratio, 1.0)) {
-		graphics_layer->setModelview(QMatrix4x4());
-	} else {
-		QMatrix4x4 matrix;
-		matrix.scale(pixelratio, pixelratio);
-		graphics_layer->setModelview(matrix);
-	}
+	painter.restore();
 
 	// Draw selection rectangle
 	if (m_selecting) {
 		fill = border = palette().color(QPalette::Highlight);
 		fill.setAlpha(48);
-		drawArray(m_selection_array, fill, border);
+		drawRect(painter, m_selection, fill, border);
 	}
 
 	// Draw message
-	m_message->draw();
-	graphics_layer->setBlended(false);
+	m_message->draw(painter);
 }
 
 //-----------------------------------------------------------------------------
 
 void Board::keyPressEvent(QKeyEvent* event)
 {
-	int offset = (event->modifiers() & Qt::ControlModifier) ? 1 : 10;
+	const int offset = (event->modifiers() & Qt::ControlModifier) ? 1 : 10;
 	switch (event->key()) {
 	// Scroll left
 	case Qt::Key_Left:
@@ -760,7 +739,7 @@ void Board::keyPressEvent(QKeyEvent* event)
 			m_action_key = event->key();
 		}
 	}
-	QOpenGLWidget::keyPressEvent(event);
+	QWidget::keyPressEvent(event);
 }
 
 //-----------------------------------------------------------------------------
@@ -770,7 +749,7 @@ void Board::keyReleaseEvent(QKeyEvent* event)
 	if (!event->isAutoRepeat()) {
 		m_action_key = 0;
 	}
-	QOpenGLWidget::keyReleaseEvent(event);
+	QWidget::keyReleaseEvent(event);
 }
 
 //-----------------------------------------------------------------------------
@@ -788,7 +767,7 @@ void Board::mousePressEvent(QMouseEvent* event)
 		m_select_pos = event->pos();
 	}
 
-	QOpenGLWidget::mousePressEvent(event);
+	QWidget::mousePressEvent(event);
 }
 
 //-----------------------------------------------------------------------------
@@ -817,28 +796,28 @@ void Board::mouseReleaseEvent(QMouseEvent* event)
 	}
 	m_action_button = Qt::NoButton;
 
-	QOpenGLWidget::mouseReleaseEvent(event);
+	QWidget::mouseReleaseEvent(event);
 }
 
 //-----------------------------------------------------------------------------
 
 void Board::mouseMoveEvent(QMouseEvent* event)
 {
-	QPoint delta = (event->pos() / m_scale) - (m_cursor_pos / m_scale);
+	const QPoint delta = (event->pos() / m_scale) - (m_cursor_pos / m_scale);
 
 	if (m_scrolling) {
 		scroll(delta);
 	}
 
 	if (!m_active_pieces.isEmpty()) {
-		int count = m_active_pieces.count();
-		for (int i = 0; i < count; ++i) {
-			m_active_pieces.at(i)->moveBy(delta);
+		// Move active pieces
+		for (Piece* piece : std::as_const(m_active_pieces)) {
+			piece->moveBy(delta);
 		}
 
 		// Attach neighbors if only one piece is active
 		if (m_active_pieces.count() == 1) {
-			m_active_pieces.first()->attachNeighbors();
+			m_active_pieces.first()->attachSolutionNeighbors();
 			updateCompleted();
 		}
 
@@ -852,7 +831,7 @@ void Board::mouseMoveEvent(QMouseEvent* event)
 		m_selecting = QVector2D(event->pos() - m_select_pos).length() >= QApplication::startDragDistance();
 	}
 	if (m_selecting) {
-		QRect rect = QRect(mapPosition(event->pos()), mapPosition(m_select_pos)).normalized();
+		const QRect rect = QRect(mapPosition(event->pos()), mapPosition(m_select_pos)).normalized();
 
 		// Check for pieces that are now selected
 		for (int i = 0; i < m_pieces.count(); ++i) {
@@ -874,7 +853,7 @@ void Board::mouseMoveEvent(QMouseEvent* event)
 			}
 		}
 
-		updateArray(m_selection_array, QRect(event->pos(), m_select_pos).normalized(), 3000);
+		m_selection = QRect(event->pos(), m_select_pos).normalized();
 	}
 
 	update();
@@ -888,6 +867,14 @@ void Board::mouseMoveEvent(QMouseEvent* event)
 
 //-----------------------------------------------------------------------------
 
+void Board::resizeEvent(QResizeEvent* event)
+{
+	updateViewport();
+	QWidget::resizeEvent(event);
+}
+
+//-----------------------------------------------------------------------------
+
 void Board::wheelEvent(QWheelEvent* event)
 {
 	if (event->angleDelta().y() > 0) {
@@ -896,7 +883,15 @@ void Board::wheelEvent(QWheelEvent* event)
 		zoomOut();
 	}
 
-	QOpenGLWidget::wheelEvent(event);
+	QWidget::wheelEvent(event);
+}
+
+//-----------------------------------------------------------------------------
+
+void Board::edgeScroll(int horizontal, int vertical)
+{
+	scroll(QPoint((5 * horizontal) / m_scale, (5 * vertical) / m_scale));
+	update();
 }
 
 //-----------------------------------------------------------------------------
@@ -920,10 +915,10 @@ void Board::stopScrolling()
 void Board::scroll(const QPoint& delta)
 {
 	m_pos -= delta;
-	int count = m_active_pieces.count();
-	for (int i = 0; i < count; ++i) {
-		m_active_pieces.at(i)->moveBy(-delta);
+	for (Piece* piece : std::as_const(m_active_pieces)) {
+		piece->moveBy(-delta);
 	}
+	updateViewport();
 }
 
 //-----------------------------------------------------------------------------
@@ -981,7 +976,6 @@ void Board::grabPiece()
 	}
 	m_active_pieces.append(piece);
 	m_pieces.removeAll(piece);
-	piece->setDepth(m_active_pieces.count() + 1);
 	piece->setSelected(true);
 	updateCursor();
 
@@ -999,22 +993,19 @@ void Board::releasePieces()
 	// Inform user this may take awhile
 	updateStatusMessage(tr("Placing pieces..."));
 	QApplication::setOverrideCursor(Qt::WaitCursor);
+	QCoreApplication::processEvents();
 
 	// Attach to closest piece
-	int count = m_active_pieces.count();
-	if (count == 1) {
-		m_active_pieces.first()->attachNeighbors();
+	if (m_active_pieces.count() == 1) {
+		m_active_pieces.first()->attachSolutionNeighbors();
 		updateCompleted();
 	}
 
 	// Place pieces
-	Piece* piece;
-	for (int i = 0; i < count; ++i) {
-		piece = m_active_pieces.at(i);
+	for (Piece* piece : std::as_const(m_active_pieces)) {
 		m_pieces.append(piece);
-		piece->setDepth(0);
 		piece->setSelected(false);
-		piece->pushNeighbors();
+		piece->pushCollidingPieces();
 	}
 	m_active_pieces.clear();
 
@@ -1047,12 +1038,11 @@ void Board::rotatePiece()
 			return;
 		}
 		piece->rotate(mapCursorPosition());
-		piece->attachNeighbors();
-		piece->pushNeighbors();
+		piece->attachSolutionNeighbors();
+		piece->pushCollidingPieces();
 	} else {
-		int count = m_active_pieces.count();
-		for (int i = 0; i < count; ++i) {
-			m_active_pieces.at(i)->rotate(mapCursorPosition());
+		for (Piece* piece : std::as_const(m_active_pieces)) {
+			piece->rotate(mapCursorPosition());
 		}
 	}
 	updateCompleted();
@@ -1070,12 +1060,8 @@ void Board::selectPieces()
 {
 	m_selecting = false;
 
-	QPoint cursor = mapCursorPosition();
-	int depth = m_active_pieces.count() + 1;
-	int count = m_selected_pieces.count();
-	for (int i = 0; i < count; ++i) {
-		Piece* piece = m_selected_pieces.at(i);
-		piece->setDepth(depth + i);
+	const QPoint cursor = mapCursorPosition();
+	for (Piece* piece : std::as_const(m_selected_pieces)) {
 		if (!piece->contains(cursor)) {
 			piece->moveBy(cursor - piece->randomPoint());
 		}
@@ -1089,19 +1075,15 @@ void Board::selectPieces()
 
 //-----------------------------------------------------------------------------
 
-void Board::drawArray(const Region& region, const QColor& fill, const QColor& border)
+void Board::drawRect(QPainter& painter, const QRect& rect, const QColor& fill, const QColor& border) const
 {
-	graphics_layer->setTextureUnits(0);
+	painter.save();
 
-	graphics_layer->setColor(fill);
-	graphics_layer->draw(region.fill);
+	painter.setPen(QPen(border, 0));
+	painter.setBrush(fill);
+	painter.drawRect(rect);
 
-	graphics_layer->setColor(border);
-	graphics_layer->draw(region.border, GL_LINE_LOOP);
-
-	graphics_layer->setColor(Qt::white);
-
-	graphics_layer->setTextureUnits(1);
+	painter.restore();
 }
 
 //-----------------------------------------------------------------------------
@@ -1120,63 +1102,37 @@ void Board::loadImage()
 	}
 	QImageReader source(Path::image(m_image_path));
 
-	// Find image size
-	QSize size = source.size();
-	const GLint max_size = graphics_layer->getMaxTextureSize() / 2;
-	if (max_size < std::max(size.width(), size.height())) {
-		size.scale(max_size, max_size, Qt::KeepAspectRatio);
-	}
+	// Set pixmap fragments to match device pixel ratio for high DPI rendering
+	const qreal pixelratio = devicePixelRatioF();
+	FragmentList::setDevicePixelRatio(pixelratio);
+
+	// Create bevel texture
+	m_bevel_pixmap = QPixmap(":/bumpmap.png");
+	m_bevel_pixmap = m_bevel_pixmap.scaled(m_bevel_pixmap.size() * pixelratio, Qt::IgnoreAspectRatio, Qt::SmoothTransformation);
+	m_bevel_pixmap.setDevicePixelRatio(pixelratio);
+
+	// Create shadow textures
+	m_shadow_pixmap = AppearanceDialog::shadow(pixelratio);
+	m_selected_shadow_pixmap = AppearanceDialog::shadowSelected(pixelratio);
 
 	// Create puzzle texture
-	int tile_size = Tile::size;
-	if (m_columns > m_rows) {
-		tile_size = std::min(tile_size, size.width() / m_columns);
-	} else {
-		tile_size = std::min(tile_size, size.height() / m_rows);
-	}
-	QSize scaled_size = size;
-	size = QSize(m_columns * tile_size, m_rows * tile_size);
+	const QSize size(m_columns * Tile::size * pixelratio, m_rows * Tile::size * pixelratio);
+	QSize scaled_size = source.size();
 	scaled_size.scale(size, Qt::KeepAspectRatioByExpanding);
 	source.setScaledSize(scaled_size);
 	source.setScaledClipRect(QRect((scaled_size.width() - size.width()) / 2, (scaled_size.height() - size.height()) / 2, size.width(), size.height()));
-	QImage image = source.read();
+	const QImage image = source.read();
 
-	int image_texture_size = powerOfTwo(std::max(size.width(), size.height()));
-	m_image_ts = static_cast<float>(tile_size) / static_cast<float>(image_texture_size);
-	QImage texture(image_texture_size, image_texture_size, QImage::Format_ARGB32);
-	texture.fill(QColor(Qt::darkGray).rgba());
+	m_pixmap = QPixmap(image.width(), image.height());
+	m_pixmap.fill(Qt::darkGray);
 	{
-		QPainter painter(&texture);
+		QPainter painter(&m_pixmap);
 		painter.drawImage(0, 0, image, 0, 0, image.width(), image.height(), Qt::AutoColor | Qt::AvoidDither);
 	}
-	m_image = new QOpenGLTexture(texture);
-	m_image->setMinificationFilter(QOpenGLTexture::LinearMipMapLinear);
-	m_image->setMagnificationFilter(QOpenGLTexture::Linear);
-	m_image->setWrapMode(QOpenGLTexture::ClampToEdge);
-
-	// Create corners
-	m_corners[0][0] = QPointF(0,0);
-	m_corners[0][1] = QPointF(0,m_image_ts);
-	m_corners[0][2] = QPointF(m_image_ts,m_image_ts);
-	m_corners[0][3] = QPointF(m_image_ts,0);
-
-	m_corners[1][0] = m_corners[0][1];
-	m_corners[1][1] = m_corners[0][2];
-	m_corners[1][2] = m_corners[0][3];
-	m_corners[1][3] = m_corners[0][0];
-
-	m_corners[2][0] = m_corners[0][2];
-	m_corners[2][1] = m_corners[0][3];
-	m_corners[2][2] = m_corners[0][0];
-	m_corners[2][3] = m_corners[0][1];
-
-	m_corners[3][0] = m_corners[0][3];
-	m_corners[3][1] = m_corners[0][0];
-	m_corners[3][2] = m_corners[0][1];
-	m_corners[3][3] = m_corners[0][2];
+	m_pixmap.setDevicePixelRatio(pixelratio);
 
 	// Create overview
-	m_overview->load(image, devicePixelRatioF());
+	m_overview->load(image, pixelratio);
 }
 
 //-----------------------------------------------------------------------------
@@ -1225,46 +1181,18 @@ QPoint Board::mapPosition(const QPoint& position) const
 
 void Board::updateCompleted()
 {
-	int t = 100 * (pieceCount() - 1);
-	int T = m_total_pieces - 1;
-	m_completed = 100 - (t / T);
+	const int pieces = 100 * (pieceCount() - 1);
+	const int total = m_total_pieces - 1;
+	m_completed = 100 - (pieces / total);
 	Q_EMIT completionChanged(m_completed);
 }
 
 //-----------------------------------------------------------------------------
 
-void Board::updateArray(Region& region, const QRect& rect, int z)
-{
-	int x1 = rect.x();
-	int y1 = rect.y();
-	int x2 = x1 + rect.width();
-	int y2 = y1 + rect.height();
-
-	graphics_layer->updateArray(region.fill,
-	{
-		Vertex::init(x1,y1,z),
-		Vertex::init(x1,y2,z),
-		Vertex::init(x2,y1,z),
-		Vertex::init(x2,y1,z),
-		Vertex::init(x1,y2,z),
-		Vertex::init(x2,y2,z)
-	});
-
-	graphics_layer->updateArray(region.border,
-	{
-		Vertex::init(x1,y1,z),
-		Vertex::init(x1,y2,z),
-		Vertex::init(x2,y2,z),
-		Vertex::init(x2,y1,z)
-	});
-}
-
-//-----------------------------------------------------------------------------
-
 void Board::updateSceneRectangle()
 {
 	m_scene = QRect(0,0,0,0);
-	for (Piece* piece : std::as_const(m_pieces)) {
+	for (const Piece* piece : std::as_const(m_pieces)) {
 		updateSceneRectangle(piece);
 	}
 }
@@ -1274,26 +1202,43 @@ void Board::updateSceneRectangle()
 void Board::updateStatusMessage(const QString& message)
 {
 	Q_EMIT showMessage(message);
+	QCoreApplication::processEvents();
 }
 
 //-----------------------------------------------------------------------------
 
-Piece* Board::pieceUnderCursor()
+void Board::updateViewport()
 {
-	QPoint pos = mapCursorPosition();
-	Piece* piece;
+	m_viewport = rect();
+
+	m_viewport_transform.reset();
+	m_viewport_transform.scale(m_scale, m_scale);
+	m_viewport_transform.translate(std::lround((width() / (2 * m_scale)) - m_pos.x()), std::lround((height() / (2 * m_scale)) - m_pos.y()));
+
+	const QRect scene = m_scene.width() ? m_viewport_transform.mapRect(m_scene) : m_viewport.adjusted(2, 2, -4, -4);
+	m_scroll_left->setVisible((scene.left() - 2) < m_viewport.left());
+	m_scroll_right->setVisible((scene.right() + 2) > m_viewport.right());
+	m_scroll_up->setVisible((scene.top() - 2) < m_viewport.top());
+	m_scroll_down->setVisible((scene.bottom() + 2) > m_viewport.bottom());
+}
+
+//-----------------------------------------------------------------------------
+
+Piece* Board::pieceUnderCursor() const
+{
+	const QPoint pos = mapCursorPosition();
 	for (int i = m_pieces.count() - 1; i >= 0; --i) {
-		piece = m_pieces.at(i);
+		Piece* piece = m_pieces.at(i);
 		if (piece->contains(pos)) {
 			return piece;
 		}
 	}
-	return 0;
+	return nullptr;
 }
 
 //-----------------------------------------------------------------------------
 
-int Board::pieceCount()
+int Board::pieceCount() const
 {
 	return m_pieces.count() + m_active_pieces.count() + m_selected_pieces.count();
 }
@@ -1321,7 +1266,6 @@ void Board::finishGame()
 
 	// Hide scene rectangle
 	m_scene = piece->boundingRect().adjusted(1,1,-2,-2);
-	updateArray(m_scene_array, m_scene, 0);
 
 	m_overview->hide();
 	unsetCursor();
@@ -1342,13 +1286,14 @@ void Board::finishGame()
 
 void Board::cleanup()
 {
-	delete m_image;
-	m_image = nullptr;
+	m_pixmap = QPixmap();
 
 	Q_EMIT clearMessage();
 	m_overview->reset();
 	m_message->setVisible(false);
+	qDeleteAll(m_active_pieces);
 	m_active_pieces.clear();
+	qDeleteAll(m_selected_pieces);
 	m_selected_pieces.clear();
 	qDeleteAll(m_pieces);
 	m_pieces.clear();
@@ -1356,14 +1301,19 @@ void Board::cleanup()
 	m_id = 0;
 	m_columns = 0;
 	m_rows = 0;
+	m_scale_level = -1;
 
 	m_scene = QRect(0,0,0,0);
-	updateArray(m_scene_array, m_scene, 0);
 	m_scrolling = false;
 	m_pos = QPoint(0, 0);
 	m_finished = false;
 
 	QSettings().remove("OpenGame");
+
+	m_scroll_left->hide();
+	m_scroll_right->hide();
+	m_scroll_up->hide();
+	m_scroll_down->hide();
 }
 
 //-----------------------------------------------------------------------------
